package net.sf.corn.exception;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Locale;
import java.util.MissingResourceException;
import java.util.ResourceBundle;
import java.util.concurrent.CopyOnWriteArrayList;

/**
 * This class is a helper to generate exceptions, prepare messages to put in
 * exceptions, take actions around it and publish events. </p>The sample code
 * below to show its basic usage: </p><code>
 *  ExceptionBuilder.with(IllegalArgumentException.class).message("Unsupported value!").execute(logError).notify(errListener).throwIt();
 * </code> </p> The code above throws IllegalArgumentException with
 * "Unsupported value!" message.It also executes logError command and notifies a
 * listener about exception throwing.Since commands and listeners are
 * predefined, they can be used in any exception throwing steps. Therefore, the
 * builder helps to shorten your codes.
 * <p>
 * The builder also helps to prepare messages from message templates:
 * </p>
 * <code>
 *  ExceptionBuilder.with(ValidationException.class).fill("Field %1! has invalid value!",field1Name).throwIt();
 *  </code> </p>Another use case of the builder is preparing messages for
 * different languages to gain i18n support for exceptions. The code below shows
 * preparing a message from a message code that defined in
 * "ExceptionMessages.properties" messages bundle. </p><code>
 *  ExceptionBuilder.with(ValidationException.class).decode("validate.invalidField",field1Name).throwIt();
 *  </code></p> Although the builder comes with a default message bundle file
 * support, it also supports other existing bundles: </p><code>
 *  ExceptionBuilder.with(ValidationException.class).bundle("com.foo.resource.Messages").decode("validate.invalidField",field1Name).throwIt();
 *  </code></p> Since the resource bundles supports language overloading, you
 * can prepare properties files for different language sets like
 * "ExceptionMessages_fr.properties". </p> The builder also offers static
 * methods to attach listeners and commands that to be processes for each
 * exception throwing. </p><code>
 *  ...</br>
 *  //Some where in your app like initializing process</br>
 *  ExceptionBuilder.addDefaultCommand(logCommand);</br>
 *  ExceptionBuilder.addCommonListener(exceptionsListener);</br>
 *  ...</br>
 *  </code></p> The defined commands and attached listener in the code above
 * works for any exception throwing by ExceptionBuilder. </p><code>
 *  //Later on in your classes</br>
 *  ExceptionBuilder.with(IllegalArgumentException.class).throwIt();</br>
 *  </code> </p> There are fundamental differences between commands and
 * listeners. Commands are executed in the same thread that you're throwing your
 * exception in the same order as you defined them. The builder first executes
 * commands that you attach during your exception building process and than
 * default commands. On the other hand actions that defined by listeners
 * executed in separate threads in parallel without affecting your execution
 * time. Executions of those threads are controlled by {@link ExceptionEventBus}
 * .
 * 
 * 
 * @author Serhat Dirik
 * @see ExceptionCommand
 * @see ExceptionListener
 * @see ExceptionEventBus
 * 
 */
public class ExceptionBuilder<E extends Exception> {

	/* STATIC PART */
	/* Constants */
	private static transient char CHAR_NEW_LINE = '\n';
	/**
	 * The default bundle name that message templates written in
	 */
	public static transient String defaultBundleName = "ExceptionMessages";

	/**
	 * List of the default commands to be executed before exception thrown
	 */
	private static transient List<ExceptionCommand> defaultCommands = new CopyOnWriteArrayList<ExceptionCommand>();
	/**
	 * List of the default listener to be triggered before exception thrown
	 */
	private static transient List<ExceptionListener> commonListeners = new CopyOnWriteArrayList<ExceptionListener>();

	/**
	 * Changes the default exception message templates bundle name which is
	 * ExceptionMessages.properties by default
	 * 
	 * @param resourceBundleName
	 *            name of the resource bundle file that in the classpath
	 * @exception MissingResourceException
	 *                if no resource bundle for the specified resource bundle
	 *                name can be found
	 */
	public static synchronized void setDefaultBundleName(
			String resourceBundleName) {
		if (resourceBundleName == null) {
			throw new IllegalArgumentException(
					"bundle name can't be a null value!");
		}
		// Check if exist
		ResourceBundle.getBundle(resourceBundleName);
		defaultBundleName = resourceBundleName;
	}

	/**
	 * Gives the default exception message templates bundle name
	 * 
	 * @return name of the bundle file in the classpath
	 */
	public static String getDefaultBundleName() {
		return defaultBundleName;
	}

	/**
	 * Gives the list of the default commands to be executed for each exception
	 * thrown
	 * 
	 * @return list of the default commands
	 */
	public static List<ExceptionCommand> getDefaultCommands() {
		return Collections.unmodifiableList(defaultCommands);
	}

	/**
	 * Tells how many default command exist
	 * 
	 * @return number of the default commands
	 */
	public static int sizeOfDefaultCommands() {
		return defaultCommands.size();
	}

	/**
	 * Tells if the given command in the defaults list
	 * 
	 * @param command
	 *            the command to be check if its in the list
	 * @return true if the command in the defaults list
	 */
	public static boolean containsDefaultCommand(ExceptionCommand command) {
		return defaultCommands.contains(command);
	}

	/**
	 * Adds a command to the defaults list, so it is executed each exception
	 * thrown
	 * 
	 * @param command
	 *            The command to be added in the list
	 * @return true if successfully added
	 */
	public static boolean addDefaultCommand(ExceptionCommand command) {
		return defaultCommands.add(command);
	}

	/**
	 * Removes a command from the defaults
	 * 
	 * @param command
	 *            the command to be removed from the list
	 * @return true if successfully removed from the defaults list
	 */
	public static boolean removeDefaultCommand(ExceptionCommand command) {
		return defaultCommands.remove(command);
	}

	/**
	 * Removes whole default commands
	 */
	public static void clearDefaultCommands() {
		defaultCommands.clear();
	}

	/**
	 * Gives the list of common exception listeners
	 * 
	 * @return unmodifiable list of common exception listeners
	 */
	public static List<ExceptionListener> getCommonListeners() {
		return Collections.unmodifiableList(commonListeners);
	}

	/**
	 * Tells how many common listener are exist
	 * 
	 * @return number of the common listeners
	 */
	public static int sizeOfCommonListeners() {
		return commonListeners.size();
	}

	/**
	 * Tells whether the given listener exist in the common listeners list or
	 * not
	 * 
	 * @param listener
	 *            the listener to be checked
	 * @return true if the given listener is defined as a common listener
	 */
	public static boolean containsCommonListener(ExceptionListener listener) {
		return commonListeners.contains(listener);
	}

	/**
	 * Puts a listener in the common listeners list
	 * 
	 * @param listener
	 *            the listener to be added in the list of common listeners
	 * @return true if succeffully added
	 */
	public static boolean addCommonListener(ExceptionListener listener) {
		return commonListeners.add(listener);
	}

	/**
	 * Removes the given listener form the common listeners list
	 * 
	 * @param listener
	 *            the listener to be removed from the commons
	 * @return true if it successfully removed
	 */
	public static boolean removeCommonListener(ExceptionListener listener) {
		return commonListeners.remove(listener);
	}

	/**
	 * Clears the whole commons list
	 */
	public static void clearCommonListeners() {
		commonListeners.clear();
	}

	/**
	 * Constructs a new Builder from the given exception class
	 * 
	 * @param exceptionClass
	 *            the class that we construct a builder for
	 * @return a new ExceptionBuilder instance
	 */
	public static <T extends Exception> ExceptionBuilder<T> with(
			Class<T> exceptionClass) {
		return new ExceptionBuilder<T>(exceptionClass);
	}

	/**
	 * Constructs a new Builder from the given exception class
	 * 
	 * @param exceptionClass
	 *            the class that we construct a builder for
	 * @param cause
	 *            Cause of the exception that we build finally
	 * @return a new ExceptionBuilder instance
	 */
	public static <T extends Exception> ExceptionBuilder<T> with(
			Class<T> exceptionClass, Throwable cause) {
		return new ExceptionBuilder<T>(exceptionClass, cause);
	}

	/* OBJECT PART */

	/**
	 * List of the exception messages
	 */
	private List<String> messages = new ArrayList<String>();

	/**
	 * List of commands to be executed before build
	 */
	private transient List<ExceptionCommand> commands = new ArrayList<ExceptionCommand>();
	/**
	 * List of the listeners that to be notified
	 */
	private transient List<ExceptionListener> listeners = new ArrayList<ExceptionListener>();
	/**
	 * Name of the bundle if common one undesired to be used
	 */
	private transient String bundleName = null;
	/**
	 * Exception class
	 */
	private transient Class<E> exceptionClass = null;

	/**
	 * Exception class default constructor
	 */
	private transient Constructor<E> exceptionDefaultConstructor = null;

	/**
	 * Exception class constructor for message only
	 */
	private transient Constructor<E> exceptionMsgOnlyConstructor = null;

	/**
	 * Exception class constructor for message & cause
	 */
	private transient Constructor<E> exceptionMsgCauseConstructor = null;

	/**
	 * Cause of the exception
	 */
	private transient Throwable cause = null;

	/**
	 * Code value for coded exceptions
	 */
	private transient Integer   code = null;
	
	/* Constructors */
	/**
	 * Creates a new builder by setting its target as the given exception class.
	 * The eligible exception classes to be used with the builder must consist a
	 * default constructor or a String parameterized constructor or both String
	 * and Throwable parameterized constructor, otherwise this constructor
	 * simply throws {@link IllegalArgument} runtime exception.
	 * 
	 * @param exceptionClass
	 *            The exception to be build
	 */
	public ExceptionBuilder(Class<E> exceptionClass) {
		this(exceptionClass, null);
	}

	/**
	 * Creates a new builder by setting its target as the given exception class
	 * and sets cause of that exception as the given cause. The eligible
	 * exception classes to be used with the builder must consist a default
	 * constructor or a String parameterized constructor or both String and
	 * Throwable parameterized constructor, otherwise this constructor simply
	 * throws {@link IllegalArgument} runtime exception.
	 * 
	 * @param exceptionClass
	 *            The exception to be build
	 * @param cause
	 *            The cause of the building exception
	 */
	public ExceptionBuilder(Class<E> exceptionClass, Throwable cause) {
		if (exceptionClass == null) {
			throw new IllegalArgumentException("Exception class can't be null!");
		}
		this.exceptionClass = exceptionClass;
		this.cause = cause;
		// Resolve constructors
		if (Modifier.isAbstract(exceptionClass.getModifiers())) {
			throw new IllegalArgumentException("Abstract exception class "
					+ exceptionClass.getName()
					+ " is not eligible to use with ExceptionBuilder!");
		}
		try {
			this.exceptionDefaultConstructor = exceptionClass.getConstructor();
			if (!Modifier.isPublic(this.exceptionDefaultConstructor
					.getModifiers())) {
				this.exceptionDefaultConstructor = null;
			}
		} catch (Exception e) {
			this.exceptionDefaultConstructor = null;
		}
		try {
			this.exceptionMsgOnlyConstructor = exceptionClass
					.getConstructor(String.class);
			if (!Modifier.isPublic(this.exceptionMsgOnlyConstructor
					.getModifiers())) {
				this.exceptionMsgOnlyConstructor = null;
			}
		} catch (Exception e) {
			this.exceptionMsgOnlyConstructor = null;

		}
		if (cause != null) {
			try {
				this.exceptionMsgCauseConstructor = exceptionClass
						.getConstructor(String.class, Throwable.class);
				if (!Modifier.isPublic(this.exceptionMsgCauseConstructor
						.getModifiers())) {
					this.exceptionMsgCauseConstructor = null;
				}
			} catch (Exception e) {
				this.exceptionMsgCauseConstructor = null;
			}
		}

		if (this.exceptionDefaultConstructor == null) {
			if ((this.exceptionMsgOnlyConstructor == null && cause == null)
					|| (this.exceptionMsgOnlyConstructor == null && this.exceptionMsgCauseConstructor == null)) {
				throw new IllegalArgumentException("Exception class "
						+ exceptionClass.getName()
						+ " is not eligible to use with ExceptionBuilder!");
			}
		}
	}

	/* Object Methods */

	/**
	 * Gives the name of the message templates bundle to be used to resolve
	 * message codes while building messages by using {@link #messageCode}
	 * method.
	 * 
	 * @return name of the bundle file
	 */
	public String getBundleName() {
		return this.bundleName == null ? defaultBundleName : this.bundleName;
	}

	/**
	 * Gives the saved messages as a multi line String
	 * 
	 * @return messages as a multi line String
	 */
	public String getMessage() {
		StringBuilder sb = new StringBuilder();
		boolean first = true;
		for (String message : messages) {
			if (!first) {
				sb.append(CHAR_NEW_LINE);
			} else {
				first = false;
			}
			sb.append(message);
		}
		return sb.length() == 0 ? null : sb.toString();
	}

	/**
	 * Gives the list of prepared messages to be used with the building
	 * exception
	 * 
	 * @return list of the prepared messages
	 */
	public List<String> getMessages() {
		return messages;
	}

	@Override
	public String toString() {
		return getMessage();
	}

	/*
	 * Builder methods
	 */
	
	/**
	 * Sets the exception code for coded exceptions. Coded exceptions are
	 * simply regular exceptions with code method to set an integer value as 
	 * the category of the exception. 
	 * 
	 * @param code an integer value for coded exceptions
	 * @return this
	 */
	public ExceptionBuilder<E> code(int code) {
		this.code = code;
		return this;
	}
	
	/**
	 * Sets the cause of the building exception
	 * 
	 * @param cause
	 *            a throwable as cause of the building exception
	 * @return this
	 */
	public ExceptionBuilder<E> cause(Throwable cause) {
		this.cause = cause;
		return this;
	}

	/**
	 * Changes the exception message templates bundle name to be used while
	 * preparing messages from the message codes
	 * 
	 * @param bundleName
	 *            name of the resource bundle file that in the classpath
	 * @exception MissingResourceException
	 *                if no resource bundle for the specified resource bundle
	 *                name can be found
	 * @return this
	 */
	public ExceptionBuilder<E> useBundle(String bundleName) {
		if (bundleName != null) {
			// Check if exist
			ResourceBundle.getBundle(bundleName);
		}
		this.bundleName = bundleName;
		return this;
	}

	/**
	 * Adds a new message to messages stack
	 * 
	 * @param message
	 *            The message to added to the messages stack
	 * @return this
	 */
	public ExceptionBuilder<E> message(String message) {
		if (message != null && message.trim().length() > 0)
			messages.add(message.trim());
		return this;
	}

	/**
	 * Builds a new message by reading the coded value from the message
	 * templates bundle and replacing parameters in the template with given
	 * message parts. Silently adds newly built message to messages stack. This
	 * method uses the systems default locale while reading the bundle.
	 * 
	 * @param messageCode
	 *            The message code is used as key to extract proper message
	 *            template from the bundle
	 * @param msgPart
	 *            Message parts are real values to be replaced with parameters
	 *            in message templates
	 * @exception MissingResourceException
	 *                if no resource bundle for the specified bundle name can be
	 *                found
	 * @return this
	 */
	public ExceptionBuilder<E> decode(String messageCode, String... msgPart) {
		return decode(messageCode, Locale.getDefault(), msgPart);
	}

	/**
	 * Builds a new message by reading the coded value from the message
	 * templates bundle and replacing parameters in the template with given
	 * message parts. Silently adds newly built message to messages stack. This
	 * method uses the given locale while reading the bundle.
	 * 
	 * @param messageCode
	 *            The message code is used as key to extract proper message
	 *            template from the bundle
	 * @param msgPart
	 *            Message parts are real values to be replaced with parameters
	 *            in message templates
	 * @exception MissingResourceException
	 *                if no resource bundle for the specified bundle name can be
	 *                found
	 * @return this
	 */
	public ExceptionBuilder<E> decode(String messageCode, Locale locale,
			String... msgPart) {
		ResourceBundle bundle = null;
		try {
			bundle = ResourceBundle.getBundle(getBundleName(), locale);
		} catch (MissingResourceException e) {
			// Try to get the default bundle
			bundle = ResourceBundle.getBundle(getBundleName());
		}
		return decode(messageCode, bundle, msgPart);
	}

	private ExceptionBuilder<E> decode(String messageCode,
			ResourceBundle bundle, String... msgPart) {
		String messageTemplate = bundle.getString(messageCode);
		if (messageTemplate == null) {
			throw new MissingResourceException("Message code " + messageCode
					+ " not found in " + getBundleName() + " !",
					String.class.getName(), getBundleName() + "." + messageCode);
		}
		return fill(messageTemplate, msgPart);
	}

	/**
	 * Builds a new message by replacing parameters in the given message
	 * template with the given message part values and adds it to the messages
	 * stack. An example could be :
	 * <p>
	 * <code>
	 *  builder.messageTemplate(" This is a %1! message for %2!","simple","you");
	 * </code>
	 * </p>
	 * 
	 * @param messageTemplate
	 *            A Message template which contains place marks as
	 *            "%1!","%2!",.. to be replaced with message parts as the same
	 *            order
	 * @param msgPart
	 *            Message part values
	 * @return this
	 */
	public ExceptionBuilder<E> fill(String messageTemplate, String... msgPart) {
		for (int i = 1; i < msgPart.length + 1; i++) {
			messageTemplate = messageTemplate.replace("%" + i + "!",
					msgPart[i - 1]);
		}
		return message(messageTemplate);
	}

	/**
	 * Adds a command to the commands stack which will be executed in order
	 * before the exception is built
	 * 
	 * @param command
	 *            command to be added to commands stack
	 * @return this
	 */
	public ExceptionBuilder<E> execute(ExceptionCommand command) {
		commands.add(command);
		return this;
	}

	/**
	 * Adds the given listener to the listeners stack which will be notified
	 * before the exception is built
	 * 
	 * @param listener
	 * @return
	 */
	public ExceptionBuilder<E> notify(ExceptionListener listener) {
		listeners.add(listener);
		return this;
	}

	/**
	 * Builds the exception
	 * 
	 * @return The prepared exception
	 */
	public E build() {
		String message = getMessage();
		E exception = null;
		try {
			if (message == null) {
				if (exceptionMsgCauseConstructor != null) {
					exception = exceptionMsgCauseConstructor.newInstance(null,
							cause);
				} else if (exceptionDefaultConstructor != null) {
					exception = exceptionDefaultConstructor.newInstance();
					exception.initCause(cause);
				} else {
					exception = exceptionMsgOnlyConstructor.newInstance("");
					exception.initCause(cause);
				}
			} else {
				if (exceptionMsgCauseConstructor != null) {
					exception = exceptionMsgCauseConstructor.newInstance(
							message, cause);
				} else if (exceptionMsgOnlyConstructor != null) {
					exception = exceptionMsgOnlyConstructor
							.newInstance(message);
					exception.initCause(cause);
				} else {
					exception = exceptionDefaultConstructor.newInstance();
					exception.initCause(cause);
				}
			}
		} catch (Exception e) {
			// Never expected to be here
			e.printStackTrace();
		}
		if(this.code != null){
			Method methodCode = null;
			try {
				methodCode = exception.getClass().getMethod("code",int.class);
			} catch (Exception e) {
			}
			if(methodCode == null){
				try {
					methodCode = exception.getClass().getMethod("code",Integer.class);
				} catch (Exception e) {
				}
			}
			if(methodCode != null && 
				( methodCode.getReturnType().equals(exception.getClass()) ||
						methodCode.getReturnType().equals(Void.class))
					){
					try {
						methodCode.invoke(exception, this.code);
					} catch (Exception e) {
						e.printStackTrace();
					}
			}
		}
		process(exception);
		return exception;
	}

	private void process(Exception e) {
		// Notify private listeners first
		for (ExceptionListener listener : listeners) {
			if (listener.isSubject(exceptionClass, true)) {
				ExceptionEventsBus.execute(listener.action(e, true));
			}
		}
		// Notify common listeners
		for (ExceptionListener listener : getCommonListeners()) {
			if (listener.isSubject(exceptionClass, true)) {
				ExceptionEventsBus.execute(listener.action(e, true));
			}
		}
		// Execute private commands
		for (ExceptionCommand command : commands) {
			try {
				command.execute(e);
			} catch (Exception ex) {
			}
		}
		// Execute default commands
		for (ExceptionCommand command : getDefaultCommands()) {
			try {
				command.execute(e);
			} catch (Exception ex) {
			}
		}
	}

	/**
	 * Throws the built exception and before that notifies listeners (private
	 * and common listener) and executes commands (private and default
	 * commands).
	 * 
	 * @throws E
	 *             The built exception
	 */
	public void throwIt() throws E {
		throw build();
	}
}
